from __future__ import division
import redpipe
from . import luascripts


def create_storage(keysp='MAB', conn='test'):
    class storage(redpipe.Hash):
        keyspace = keysp
        connection = conn

    return storage


class ThompsonSamplingMultiArmedBandit(object):
    def __init__(self, name, arms, storage=None, alpha=5, beta=5,
                 expires=None):

        self.name = name
        self.alpha = alpha
        self.beta = beta
        self.arms = sorted(set(arms))
        self.expires = expires
        if storage is None:
            storage = create_storage()

        self.storage = storage

    @classmethod
    def _pipe(cls, pipe=None, autoexec=True):
        return redpipe.pipeline(pipe=pipe, autoexec=autoexec)

    def delete(self, pipe=None):
        with self._pipe(pipe=pipe) as p:
            s = self.storage(pipe=p)
            s.delete(self.name)

    def draw(self, pipe=None):
        with self._pipe(pipe=pipe) as p:
            s = self.storage(pipe=p)
            response = redpipe.Future()
            result = s.eval(luascripts.draw_lua, 1, self.name, self.alpha,
                            self.beta, *[a for a in self.arms])
            if self.expires:
                s.expire(self.name, self.expires)

            def cb():
                response.set(s.valueparse.decode(result))

            p.on_execute(cb)

            return response

    def draw_multi(self, times, pipe=None):
        with self._pipe(pipe=pipe) as p:
            return [self.draw(pipe=p) for _ in range(times)]

    def update_sucess(self, arm, reward=1.0, pipe=None):
        with self._pipe(pipe=pipe) as p:
            s = self.storage(pipe=p)
            s.eval(luascripts.update_success_lua, 1, self.name, arm, reward,
                   self.alpha, self.beta)
            if self.expires:
                s.expire(self.name, self.expires)

    def state(self, pipe=None):
        with self._pipe(pipe=pipe) as p:
            return self.storage(pipe=p).hgetall(self.name)



draw_lua = """
local name = KEYS[1]
local alpha = tonumber(ARGV[1])
local beta = tonumber(ARGV[2])

local arms = {}

for i=3,#ARGV do
    arms[#arms + 1] = ARGV[i]
end

local max_mean = 0
local mean = 0
local count = 0
local arm = arms[1]
local bulk = redis.call('HGETALL', name)
local result = {}
local k, i, v

local beta_mean = function(success, count, alpha, beta)
    return 1 / (1 + (count - success + beta) / (success + alpha))
end

local init_mean = beta_mean(0, 0, alpha, beta)

for i, v in ipairs(bulk) do
    if i % 2 == 1 then
        k = v
    else
        result[k] = v
    end
end

for i, a in ipairs(arms) do
    mean = tonumber(result["#{" .. a .. "}:mean"] or init_mean)
    if mean > max_mean then
        max_mean = mean
        arm = a
    end
end

local count = tonumber(result["#{" .. arm .. "}:count"] or 0) + 1
local success = tonumber(result["#{" .. arm .. "}:success"] or 0)
redis.call('HSET', name, "#{" .. arm .. "}:count", count)
mean = beta_mean(success, count, alpha, beta)
redis.call('HSET', name, "#{" .. arm .. "}:mean", mean)
return arm
"""

update_success_lua = """
local name = KEYS[1]
local arm = ARGV[1]
local reward = tonumber(ARGV[2])
local alpha = tonumber(ARGV[3])
local beta = tonumber(ARGV[4])
local success = tonumber(
    redis.call('HINCRBYFLOAT', name, "#{" .. arm .. "}:success", reward))
local count = tonumber(redis.call('HGET', name, "#{" .. arm .. "}:count"))
local mean = 1 / (1 + (count - success + beta) / (success + alpha))
redis.call('HSET', name, "#{" .. arm .. "}:mean", mean)
"""



# -*- coding: utf-8 -*-
"""
Utility for grabbing the redmab version string from the VERSION file.
"""

import os

__all__ = ['__version__']

DIR = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(DIR, 'VERSION')) as f:
    __version__ = str(f.read().strip())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from os import path
from setuptools import setup
from distutils.cmd import Command

NAME = 'redmab'

ROOTDIR = path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOTDIR, 'README.rst')) as f:
    readme = f.read()

with open(os.path.join(ROOTDIR, 'docs', 'release-notes.rst')) as f:
    history = f.read()

with open(os.path.join(ROOTDIR, NAME, 'VERSION')) as f:
    version = str(f.read().strip())


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys
        import subprocess

        raise SystemExit(
            subprocess.call([sys.executable, '-m', 'test']))


cmdclass = {'test': TestCommand}
ext_modules = []

setup(
    name=NAME,
    version=version,
    description='Multi-Armed Bandit implementation in redis',
    author='John Loehrer',
    author_email='72squared@gmail.com',
    url='https://github.com/72squared/%s' % NAME,
    download_url='https://github.com/72squared/%s/archive/%s.tar.gz' %
                 (NAME, version),
    keywords='redis multi-armed bandit',
    packages=[NAME],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Environment :: Web Environment',
        'Operating System :: POSIX'],
    license='MIT',
    install_requires=['redpipe>=2.0.0'],
    tests_require=['redislite>=3.0.271', 'tox', 'coverage'],
    include_package_data=True,
    long_description=readme + '\n\n' + history,
    cmdclass=cmdclass,
    ext_modules=ext_modules
)
